import os
import re
import uuid
import logging

import jwt

from .model import ROLE_GUEST, ROLE_USER, ROLE_STAFF, User

JWT_SECRET = os.environ.get('ALLOC_JWT_SECRET', 'DEBUG_JWT_SECRET')
STAFF_PASSWORD = os.environ.get('ALLOC_STAFF_PW', 'PASSWORD')

_usercache = {}

class UserManager:
    """Container for managing users"""

    def __init__(self):
        self.cache = {}

    def exists(self, uid):
        return uid in self.cache

    def setup(self, data):
        tmp = User( data['uid'] )
        tmp.name = data.get('name', 'Unknown User')
        tmp.role = int( data.get('role', ROLE_GUEST) )
        self.cache[ str(tmp.uid) ] = tmp
        return tmp

    def get(self, uid):
        return self.cache.get( str(uid), None )

    def get_name(self, uid):
        user = self.cache.get( str(uid), None )
        if not user:
            return "Unknown User"
        return user.name

    def get_sessions(self, uid):
        user = self.cache.get( str(uid), None )
        if not user:
            return set()
        return user.sessions

def check_password( provided ) -> bool:
    """Check if a password is valid"""
    if not provided:
        return False
    return provided == STAFF_PASSWORD

def encode_session( data ):
    """Encode a session into a session token"""
    return jwt.encode( data, JWT_SECRET, algorithm="HS256" )

def decode_session( encoded ):
    """Restore a session based on a session token"""
    try:
        return jwt.decode( encoded, JWT_SECRET, algorithms=["HS256"] )
    except jwt.exceptions.InvalidSignatureError:
        return None

def setup(users: UserManager, message):
    """Setup a user account based on a user request"""
    role = ROLE_USER

    # check if the user provided a password
    if 'password' in message:
        if check_password( message['password'] ):
            role = ROLE_STAFF
        else:
            return None, None

    data = {
        'uid': str(uuid.uuid4()),
        'name': re.sub(r'[^\w -]', '', message.get('name', '')),
        'role': role
    }

    user = users.setup(data)
    ticket = encode_session(data)
    return user, ticket

def login(users: UserManager, message):
    """Handle a login message"""
    ticket = decode_session( message )
    if not ticket or 'uid' not in ticket:
        return None

    user = users.get( ticket['uid'] )
    if not user:
        logging.warning("Ticket has a valid UID, but it does not exist - restarted?")
        user = users.setup( ticket )
    return user
