
import random
import itertools
import logging
from collections import defaultdict
from typing import Optional

def lMin( items ):
    idx = 0
    val = items[0]
    for i in range(1, len(items)):
        if val < items[i]:
            idx = i
            val = items[i]
    return idx

class GroupManager:
    """Class that contains the list of groups"""

    def __init__(self):
        self.groups = []
        self.method = None
        self.max_groups = None

    def _allocate_variable(self, students, ideal_size=2, min_members=1):
        """Create groups based on an ideal size.

        The number of groups is variable."""
        if not students:
            return []

        random.shuffle( students )

        # group allocations
        groups = [ students[n:n+ideal_size] for n in range(0, len(students), ideal_size) ]

        # less than idealSize students - 1 group
        if not groups:
            return [ students ]

        n_groups = len(groups)

        # deal with the last group
        if len( groups[-1] ) < min_members and len(groups) < 1:
            last_group = groups.pop()
            n_groups -= 1
            self._allocate_uniform( last_group, n_groups )
        return groups

    def _allocate_fixed(self, students, ideal_groups=4, min_members=2):
        """Create a fixed number of groups and allocate students accordingly"""
        max_possible_groups = len( students ) // min_members
        num_groups = min( ideal_groups, max_possible_groups )

        # there must be at least one group
        if num_groups == 0:
            num_groups = 1

        students_per_group = len(students) // num_groups
        return self._allocate_variable( students, students_per_group, min_members)

    def _allocate_uniform(self, people, num_groups=None):
        people = list( people )
        if num_groups is None:
            num_groups = len( self.groups )

        group_id = 0
        while people:
            self.groups[ group_id ].append( people.pop() )
            group_id = ( group_id + 1) % num_groups
        return people

    def _allocate_smallest(self, people):
        while people:
            smallest_group = lMin( self.groups )
            self.groups[smallest_group].append( people.pop() )

    @property
    def everyone(self):
        """Get all users"""
        return set( itertools.chain( self.groups ) )

    def members(self, group):
        """Return all members of a group"""
        try:
            print( self.groups )
            return self.groups[ group ]
        except IndexError:
            return []

    def find(self, member) -> Optional[int]:
        """Find what group a member belongs to"""
        for i, group in enumerate(self.groups):
            if member in group:
                return i
        return None

    def allocate(self, students, method, ideal_number, min_members=2):
        """Do a full reallocation, ignoring any current team allocations"""
        if method == 'members':
            self.method = method
            self.groups = self._allocate_variable(students, ideal_number, min_members)
            return self.groups
        if method == 'groups':
            self.method = method
            self.groups = self._allocate_fixed(students, ideal_number, min_members)
            return self.groups

        logging.warning("got unknown group allocation method: %s", method)
        return []

    def get_group(self, uid):
        """Get a group via user id"""
        for idx, group in enumerate(self.groups):
            if uid in group:
                return idx
        return None

    def remove(self, people, min_members):
        """Remove missing members from group allocation"""
        extra_spaces = defaultdict(int)
        illegal_groups = set()
        empty_groups = set()

        for person in people:
            group_id = self.find( person )
            self.groups[ group_id ].remove( person )
            extra_spaces[ group_id ] += 1

            if len( self.groups[group_id] ) == 0:
                empty_groups.add( group_id )
            elif len( self.groups[group_id] ) < min_members:
                illegal_groups.add( group_id )

        return extra_spaces, illegal_groups, empty_groups

    def update(self, students, method):
        """Do a partial reallocation, taking into account existing teams"""
        new_people = list(set( students ) - self.everyone)
        missing_people = list(self.everyone - set(students))

        if missing_people:
            spaces, illegal, empty = self.remove( missing_people, 1 )

        if method == 'uniform':
            self._allocate_uniform( new_people )
        elif method == 'smallest':
            self._allocate_smallest( new_people )


