##
#
##

import logging
import uuid

from .model import ROLE_USER, ROLE_STAFF
from .stream import SocketServer, ClientHandler
from .util import group_json_many, group_json_one, decode_allocate


def handle_allocate( server: SocketServer, client: ClientHandler, message ):
    """Handle the allocation process"""
    if client.role != ROLE_STAFF:
        logging.warning("%s tried to use allocate, but is not staff", client.id)
        return

    # parameters
    ideal_size, min_size, method = decode_allocate( message )

    # allocation
    active_sessions = server.active( lambda x: x.role == ROLE_USER )
    users = list({ server.clients[session].user.uid for session in active_sessions })
    teams = server.groups.allocate( users, method, ideal_size, min_size )

    # figure out the full team allocation json object
    all_teams = group_json_many( server, teams )
    server.multicast_topic( server.online_staff, 'groups', all_teams )

    # send team allocations
    for team_id, team in enumerate(teams):
        payload = all_teams[ team_id ]

        # get all sessions for user teams
        sessions = set()
        for member in team:
            sessions = sessions.union( server.users.get_sessions( member ) )

        # send them the group members
        server.multicast_topic( sessions, 'group', payload )


def handle_notify( server: SocketServer, client: ClientHandler, message ):
    if not client.user or not client.user.has_access( ROLE_STAFF ):
        return

    target = server.users.get( message['data']['uuid'] )
    if not target:
        logging.warning("no idea who that is...")
        return

    if target.sessions:
        server.multicast_topic( target.sessions, 'notify', {
                'message': message['data']['message']
        })


def handle_notify_group( server: SocketServer, client: ClientHandler, message ):
    if not client.user or not client.user.has_access( ROLE_STAFF ):
        return

    group = server.groups.members( int(message['data']['group']) )
    if not group:
        logging.warning("tried to set data to non-existant group")
        return

    sessions = []
    for user_id in group:
        user = server.users.get( user_id )
        if user:
            sessions += user.sessions

    server.multicast_topic(sessions, 'notify', {
        'message': message['data']['message']
    })


def handle_kick( server: SocketServer, client: ClientHandler, message ):
    """Note this doesn't actually kick the client, it just asks it to disconnect.
    I could actually force disconnect, blacklist the session and user IDs,
    but that's a little overkill.

    It's mostly useful for force-invalidating sessions."""
    if not client.user or not client.user.has_access( ROLE_STAFF ):
        return

    target = server.users.get( message['data']['uuid'] )
    if not target:
        logging.warning("got unknown user ID")
        return

    if target.sessions:
        server.multicast_topic( target.sessions, 'kick', {} )



def handle_rename(server: SocketServer, client: ClientHandler, message):
    if not client.user or not client.user.has_access(ROLE_STAFF):
        logging.warning("non-admin tried to rename?")
        return

    if 'uuid' not in message['data'] or 'name' not in message['data']:
        logging.warning("malformed force-rename message: %s", message)
        return

    user = None
    try:
        user_id = uuid.UUID( "{{%s}}" % ( message['data']['uuid'], ) )
        user = server.users.get( user_id )
    except ValueError:
        logging.warning("Could not decode uuid")
        return

    if user is None:
        logging.warning("Staff member tried to rename unknown user")
        return

    # update the name
    user.set_name( message['data'].get('name', 'RENAMED USER') )

    # inform people about the rename
    payload = {'uid': user.id_str, 'name': user.name}
    server.multicast_topic(user.sessions, 'userRename', payload)
    server.multicast_topic(server.online_staff, 'userRenameAdm', payload)

    # reissue group data
    group = server.groups.get_group( user.uid )
    if group is not None:
        sessions = []
        members = server.groups.members( group )
        for user in members:
            sessions += server.users.get( user ).sessions
        server.multicast_topic(sessions, 'group', group_json_one( server, group, members ))
