import uuid
import json
import asyncio
import logging
from typing import Any, Callable

import websockets

from . import auth, groups, model, util

async def send(websocket, message):
    """send a message, don't care if failed"""
    try:
        await websocket.send( message )
    except websocket.ConnectionClosed:
        pass

class ClientHandler:
    """Websocket wrapper for the client"""

    def __init__(self, socket):
        self.id_ = uuid.uuid4()
        self.socket = socket
        self.user = None
        self.queue = asyncio.Queue()

    @property
    def name(self):
        """The current user's known as name"""
        if not self.user:
            return "Unknown User"
        return self.user.name

    @property
    def role(self):
        """The current users role"""
        if not self.user:
            return model.ROLE_GUEST
        return self.user.role

    async def send_now(self, topic, message):
        """Send a message to this client, bypassing the sending queue"""
        msg = {
            'type': topic,
            'data': message
        }
        await self.socket.send( json.dumps( msg ) )

    def send(self, message):
        """Queue a message to this client"""
        self.queue.put_nowait( message )

    async def _terminate(self, reason=None):
        self.socket.close(1011, reason)

    async def _login_handshake(self, server):
        """Process the login handshake with the client"""
        # start with the welcome event
        await self.send_now( 'welcome', {
            'id': self.id
        })

        # for us to get the response to the welcome event
        msg = json.loads(await self.socket.recv())
        if 'type' not in msg:
            return None, None

        logging.info( "%s", msg )

        # This is ticket-based auth, they previouslly used the tool
        if msg['type'] == 'ticket':
            logging.info("ticket authentication")
            user = auth.login(server.users, msg['data'])
            if not user:
                await self._terminate("invalid ticket")
                return None, None
            return user, None

        # This is a setup event, they are a new user
        if msg['type'] == 'setup':
            logging.info("user setup")
            user, ticket = auth.setup(server.users, msg['data'])
            if not ticket:
                await self._terminate("authentication failed")
                return None, None
            return user, ticket

        # I don't know what that was...
        logging.warning("invalid handshake, got message type %s", msg['type'])
        await self._terminate("unknown handshake type")
        return None, None

    async def login(self, server):
        """Login routine/handshake for the client"""

        try:
            # perform handshake
            user, ticket = await self._login_handshake(server)
            if user is None:
                return None

            # handshake is over, we know who the user is, lets process it.
            self.user = user
            group = server.groups.find( user.uid )

            # tell the user they are setup now
            await self.send_now( 'user', {
                'uid': user.uid,
                'name': user.name,
                'role': user.role,
                'group': server.groups.find( user.uid ),
                'ticket': ticket
            })

            # If this user is allocated, send the team members to the client
            if group is not None:
                await self.send_now('group', util.group_json_one( server, group,
                                                                server.groups.members( group )
                                                                ))

            # staff users get told about current users and groups
            if user.has_access( model.ROLE_STAFF ):
                online_users = [ util.status_payload(c) for c in server.clients.values() ]
                await self.send_now( 'users', online_users )
                online_groups = util.group_json_many( server, server.groups.groups )
                await self.send_now( 'groups', online_groups )

            return user
        except websockets.exceptions.ConnectionClosed:
            return None

    async def consumer(self):
        """Async worker: server -> client"""
        while True:
            message = await self.queue.get()
            logging.info("[>>] %s %s", self.id, message)
            await send( self.socket, json.dumps(message) )

    async def producer(self, server):
        """Async worker: client -> server"""
        async for message_txt in self.socket:
            message = json.loads( message_txt )
            logging.info("[<<] %s %s", self.id, message)
            await server.process(self, message)

    async def handler(self, server):
        """Client handler"""
        consumer_task = asyncio.create_task( self.consumer() )
        producer_task = asyncio.create_task( self.producer( server ) )
        done, pending = await asyncio.wait(
            [consumer_task, producer_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in pending:
            task.cancel()

    @property
    def id(self):
        """Return the ID of the user account"""
        return str(self.id_)

class SocketServer:
    """The server socket used for websocket connections."""

    def __init__(self, callbacks):
        self.clients = {}
        self.groups = groups.GroupManager()
        self.users = auth.UserManager()
        self.protocol: dict[str, Callable[[SocketServer, ClientHandler, dict[Any]], None]] =\
            callbacks

    async def handler(self, client_socket):
        client = ClientHandler(client_socket)

        # Handshake
        user = await client.login(self)
        if not user:
            return

        # main loop
        self.clients[ client.id_ ] = client
        client.user.sessions.add( client.id_ )
        self.multicast_topic( self.online_staff, 'status', util.status_payload( client ))
        await client.handler( self )
        client.user.sessions.remove( client.id_ )
        del self.clients[ client.id_ ]
        self.multicast_topic( self.online_staff, 'status', util.status_payload(client, False) )

    async def process(self, client, message):
        if 'type' not in message:
            return

        topic = message.get('type', None)
        if topic in self.protocol:
            self.protocol[ topic ]( self, client, message )

    @property
    def online_staff(self):
        """Return all staff accounts which are currently online"""
        return [ x.id_ for x in self.clients.values() if x.user.role == model.ROLE_STAFF ]

    def active_ids(self, predicate=None) -> list[str]:
        """Return active session UUIDs (as strings)"""
        if predicate is None:
            return [ str(k) for k in self.clients ]
        return [ v.id for (k,v) in self.clients.items() if predicate(v) ]

    def active(self, predicate=None) -> list[uuid.UUID]:
        """Return active session UUIDs (as UUIDs)"""
        if predicate is None:
            return list(self.clients)
        return [ v.id_ for (k,v) in self.clients.items() if predicate(v) ]

    def send_topic(self, client, topic, payload):
        """Send a topic and payload to a single client"""
        self._send(client, {
            'type': topic,
            'data': payload
        })

    def multicast_topic(self, clients, topic, payload):
        """Send a topic and payload to multiple clients"""
        self._multicast(clients, {
            'type': topic,
            'data': payload
        })

    def broadcast_topic(self, topic, payload):
        """Sent a topic and payload to all clients"""
        self._broadcast({
            'type': topic,
            'data': payload
        })

    def _send(self, client, message):
        """Send a raw message to a client"""
        self.clients[ client ].send( message )

    def _multicast(self, clients, message):
        """Send a raw message to multiple clients"""
        for cid in clients:
            if cid in self.clients:
                websocket = self.clients[ cid ]
                websocket.send( message )

    def _broadcast(self, message):
        for websocket in self.clients.values():
            websocket.send( message )
